#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include "browse.h"

static int strcasecmp_void(const void *s1, const void*s2)
{
  return strcasecmp((const char *) s1, (const char *) s2);
}

static void free_fileinfo_void(void *p)
{
  free_fileinfo((struct fileinfo*) p);
}

static void free_noeffect(void *p)
{
  p = p;
}

#define MAXBUF 4096

static int isusedby(const struct fileinfo* fi1,
		    const struct fileinfo* fi2,
		    const char* skip)
{
  FILE	*f;
  char	c;
  int	k;
  int	l;
  int	i;
  int	skip_len;

  if (!(f = fopen(fi1->absfn, "r")))
    {
      fprintf(stderr, "Cannot open file %s: %s\n", fi1->relfn, strerror(errno));
      exit(2);
    }

  l = strlen(fi2->bn_noext);
  skip_len = strlen(skip);
  k = -1;
  while ((c = fgetc(f)) != EOF)
    {
      if (k == -1)
	for (i = 0; i < skip_len; ++i)
	  if (c == skip[i])
	    {
	      while (c != '\n' && c != EOF)
		c = fgetc(f);
	      continue;
	    }
      if (0 <= k && k < l && c == fi2->bn_noext[k])
	++k;
      else
	if ((c == '.' || c == '}' || c == ']' || c == ',' || c == '"') && k == l)
	  break;
	else
	  if (c == '/' || c == '{' || c == '[' || c == ',' || c == '"')
	    k = 0;
	  else
	    k = -1;
    }
  fclose(f);
  /*fprintf(stderr, "%s %s %d %d %d\n", fi1->absfn, fi2->bn, k, l, k == l);*/
  return k == l;
}

static int searchfor(const struct fileinfo* fi1,
		     const char* s)
{
  FILE	*f;
  char	c;
  int	k;
  int   l;

  if (!(f = fopen(fi1->absfn, "r")))
    {
      fprintf(stderr, "Cannot open file %s: %s\n", fi1->relfn, strerror(errno));
      exit(2);
    }

  l = strlen(s);
  k = -1;
  while ((c = fgetc(f)) != EOF)
    {
      if (0 <= k && k < l && c == s[k])
	++k;
      else
	if (k == l)
	  break;
	else
	  if (c == s[0])
	    k = 1;
	  else
	    k = -1;
    }
  fclose(f);
  return k == l;
}

static int containspace(const char *s)
{
  for (; *s != '\0' && *s != ' '; ++s)
    ;
  return *s == ' ';
}

#define UNUSED ((void*) 0x0)
#define USED ((void*) 0x1)

int main(int argc, char* argv[])
{
  char*		 src;
  char*		 srcimagesdir;
  char*		 dst;
  char*		 format;
  struct s_list* list_filter;
  struct s_list* list;
  struct s_list* list_images;
  struct s_list* lext;
  struct s_list* p;
  struct s_list* p1;
  struct s_list* p2;
  struct fileinfo* fi;
  struct fileinfo* fi1;
  struct fileinfo* fi2;
  int are_there_unused;

  if (argc != 5)
    {
      fprintf(stderr, "Usage: %s srcdir srcimagesdir dstdir format\n", argv[0]);
      exit(2);
    }
  src = argv[1];
  srcimagesdir = argv[2];
  dst = argv[3];
  format = argv[4];

  if (src[strlen(src)-1] == '/')
    src[strlen(src)-1] = '\0';
  if (dst[strlen(dst)-1] == '/')
    dst[strlen(dst)-1] = '\0';

  list = browse(argv[0], src);
  list_images = browse(argv[0], srcimagesdir);

  /* Keep .tex .sty and .bib that have no space in filename */
  list_filter = NULL;
  for (p = list; p; p = p->next)
    {
      fi = p->data;
      if (!strcasecmp(fi->ext, ".tex") ||
	  !strcasecmp(fi->ext, ".sty") ||
	  !strcasecmp(fi->ext, ".bib"))
	{
	  if (!containspace(fi->absfn))
	    list_add_item_head(&list_filter, fi);
	  else
	    {
	      fprintf(stderr, "Cannot deal with spaces in filename \"%s\". Skipped\n", fi->relfn);
	      free_fileinfo_void(fi);
	    }
	}
      else
	free_fileinfo_void(fi);
    }
  list_free(&list, free_noeffect);
  list = list_filter;

  /* Keep images that have no spaces in filename */
  list_filter = NULL;
  for (p = list_images; p; p = p->next)
    {
      fi = p->data;
      if (!containspace(fi->absfn))
	list_add_item_head(&list_filter, fi);
      else
	{
	  fprintf(stderr, "Cannot deal with spaces in image filename \"%s\". Skipped\n", fi->relfn);
	  free_fileinfo_void(fi);
	}
    }
  list_free(&list_images, free_noeffect);
  list_images = list_filter;

  /* Extract extensions used for rules */
  lext = NULL;
  for (p = list; p; p = p->next)
    {
      fi = p->data;
      list_add_item_sorted(&lext, fi->ext, strcasecmp_void);
    }
  for (p = list_images; p; p = p->next)
    {
      fi = p->data;
      list_add_item_sorted(&lext, fi->ext, strcasecmp_void);
    }

  /* Initialized all files to unused */
  for (p = list; p; p = p->next)
    {
      fi = p->data;
      fi->extra = UNUSED;
    }
  for (p = list_images; p; p = p->next)
    {
      fi = p->data;
      fi->extra = UNUSED;
    }

  /* Create suffixes list */
  if (lext)
    {
      printf(".SUFFIXES: .pdf");
      if (!strcasecmp(format, "eps"))
	printf(" .dvi .ps .eps");
	for (p = lext; p; p = p->next)
	if (strcasecmp((char *) p->data, ""))
	  printf(" %s", (char *) p->data);
      printf("\n\n");
    }
  if (list)
    {
      printf("target:");
      for (p = list; p; p = p->next)
	{
	  fi = p->data;
	  if (!strcasecmp(fi->ext, ".tex"))
	    if (searchfor(fi, "\\documentclass") &&
		(!strcmp(fi->dn, src) || !strcmp(fi->reldn, src)))
	      {
		if (!strcasecmp(format, "pdf"))
		  printf(" %s.pdf", fi->relfn_noext);
		if (!strcasecmp(format, "eps"))
		  printf(" %s.dvi %s.ps %s.pdf",
			 fi->relfn_noext,
			 fi->relfn_noext,
			 fi->relfn_noext);
	      }
	}
      printf("\n");
      printf("\t@echo \"Compilation success\" | $(COLORIZEGREEN);\n");
      printf("\t@printf \"\\n\";\n");
      printf("\n");
    }
  for (p = list; p; p = p->next)
    {
      fi = p->data;
      if (!strcasecmp(fi->ext, ".tex"))
	if (searchfor(fi, "\\documentclass") &&
	    (!strcmp(fi->dn, src) || !strcmp(fi->reldn, src)))
	  {
	    fi->extra = USED;
	    if (!strcasecmp(format, "pdf"))
	      printf("%s.pdf: %s/%s.tex\n\n", fi->relfn_noext, dst, fi->relfn_noext);
	    if (!strcasecmp(format, "eps"))
	      printf("%s.dvi: %s/%s.tex\n\n", fi->relfn_noext, dst, fi->relfn_noext);
	  }
    }
  for (p1 = list; p1; p1 = p1->next)
    {
      fi1 = p1->data;
      if (!strcasecmp(fi1->ext, ".tex") || !strcasecmp(fi1->ext, ".sty"))
	{
	  printf("%s/%s: %s", dst, fi1->relfn, fi1->relfn);
	  for (p2 = list; p2; p2 = p2->next)
	    {
	      fi2 = p2->data;
	      if (!strcmp(fi1->relfn, fi2->relfn))
		continue;
	      if (!strcasecmp(fi2->ext, ".tex") || !strcasecmp(fi2->ext, ".sty"))
		if (isusedby(fi1, fi2, "%"))
		  {
		    fi2->extra = USED;
		    printf(" %s/%s", dst, fi2->relfn);
		  }
	      if (!strcasecmp(fi2->ext, ".bib"))
		if (isusedby(fi1, fi2, "%"))
		  {
		    fi2->extra = USED;
		    printf(" %s", fi2->relfn);
		  }
	    }
	  for (p2 = list_images; p2; p2 = p2->next)
	    {
	      struct fileinfo * fi2 = p2->data;
	      if (isusedby(fi1, fi2, "%"))
		{
		  fi2->extra = USED;
		  printf(" $(IMAGES_DST_PATH)/%s.%s", fi2->relfn_noext, format);
		}
	    }
	  printf("\n");
	  printf("\t@mkdir -p %s%s\n", dst, fi1->reldn);
	  printf("\t@cp -f %s %s/%s\n\n", fi1->relfn, dst, fi1->relfn);
	}
    }
  for (p1 = list_images; p1; p1 = p1->next)
    {
      fi1 = p1->data;
      printf("$(IMAGES_DST_PATH)/%s.%s: $(IMAGES_SRC_PATH)/%s", fi1->relfn_noext, format, fi1->relfn);
      if (!strcasecmp(fi1->ext, ".svg"))
	for (p2 = list_images; p2; p2 = p2->next)
	  {
	    fi2 = p2->data;
	    if (!strcmp(fi1->relfn, fi2->relfn))
	      continue;
	    if (isusedby(fi1, fi2, ""))
	      {
		if (!fi2->extra)
		  fi2->extra = fi1->extra;
		printf(" $(IMAGES_SRC_PATH)/%s", fi2->relfn);
	      }
	  }
      printf("\n");
      printf("\t@echo \"Generation of image %s.%s\"\n", fi1->relfn_noext, format);
      printf("\t@mkdir -p '$(IMAGES_DST_PATH)/%s'\n", fi1->reldn);
      printf("\t@$(MT_SCRIPTS_PATH)/mt_convert $(MAKEFILE_RULES) $(IMAGES_SRC_PATH)/%s $(IMAGES_DST_PATH)/%s.%s 2>&1 | $(INDENT) | ($(COLORIZEBLUE); true) \n",
	     fi1->relfn, fi1->relfn_noext, format);
      printf("\n");
    }
  if (list)
    printf("\n\n");

  for (p = list; p; p = p->next)
    {
      fi = p->data;
      if (!fi->extra && (!strcasecmp(fi->ext, ".tex") || !strcasecmp(fi->ext, ".sty") || !strcasecmp(fi->ext, ".bib")))
	fprintf(stderr, "Unused file %s\n", fi->relfn);
    }
  printf("vacuum:\n");
  printf("\t@printf 'Vacuum cleaning\\n' | $(COLORIZERED);\n");
  printf("\t@printf '===============\\n\\n' | $(COLORIZERED);\n");
  are_there_unused = 0;
  for (p = list_images; p; p = p->next)
    {
      fi = p->data;
      if (!fi->extra)
	{
	  printf("\t@echo Move $(IMAGES_SRC_PATH)/%s to $(IMAGES_TRASH_PATH) ;\n", fi->relfn);
	  printf("\t@mkdir -p $(IMAGES_TRASH_PATH)/%s ;\n", fi->reldn);
	  printf("\t@mv -f $(IMAGES_SRC_PATH)/%s $(IMAGES_TRASH_PATH)/%s ;\n", fi->relfn, fi->reldn);
	  fprintf(stderr, "Unused image %s\n", fi->relfn);
	  are_there_unused = 1;
	}
    }
  printf("\t@printf 'Vacuuming success\\n\\n' | $(COLORIZEGREEN);\n");
  if (are_there_unused)
    fprintf(stderr, "Type 'moosetex vacuum' to move the useless images to the trash directory\n");

  list_free(&list_images, free_fileinfo_void);
  list_free(&list, free_fileinfo_void);
  list_free(&lext, free_noeffect);

  return EXIT_SUCCESS;
}
